/* 
   ipmi-lan-interface-udm.c - IPMI UDM LAN Interface

   Copyright (C) 2003, 2004, 2005 FreeIPMI Core Team

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.

*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <sys/types.h>

#include <stdio.h>
#include <stdlib.h>
#ifdef STDC_HEADERS
#include <string.h>
#endif /* STDC_HEADERS */
#include <sys/socket.h>
#include <netinet/in.h>
#if TIME_WITH_SYS_TIME
#include <sys/time.h>
#include <time.h>
#else /* !TIME_WITH_SYS_TIME */
#if HAVE_SYS_TIME_H
#include <sys/time.h>
#else /* !HAVE_SYS_TIME_H */
#include <time.h>
#endif /* !HAVE_SYS_TIME_H */
#endif  /* !TIME_WITH_SYS_TIME */
#if HAVE_UNISTD_H
#include <unistd.h>
#endif /* HAVE_UNISTD_H */
#include <assert.h>
#include <errno.h>

#include "freeipmi/udm/ipmi-lan-interface-udm.h"
#include "freeipmi/ipmi-authentication-type-spec.h"
#include "freeipmi/ipmi-debug.h"
#include "freeipmi/ipmi-lan.h"
#include "freeipmi/ipmi-lan-interface.h"
#include "freeipmi/ipmi-lan-utils.h"
#include "freeipmi/rmcp.h"

#include "ipmi-udm-device.h"

#include "err-wrappers.h"
#include "fiid-wrappers.h"
#include "freeipmi-portability.h"

fiid_template_t tmpl_lan_raw_hdr = 
  {
    {2, "lun", FIID_FIELD_REQUIRED | FIID_FIELD_LENGTH_FIXED},
    {6, "net_fn", FIID_FIELD_REQUIRED | FIID_FIELD_LENGTH_FIXED},
    {0, "", 0}
  };

fiid_template_t tmpl_lan_raw =
  {
    {8192, "raw_data", FIID_FIELD_OPTIONAL | FIID_FIELD_LENGTH_VARIABLE},
    {0, "", 0}
  };

#define IPMI_LAN_RQ_SEQ_INC(__rq_seq) ((__rq_seq) = (((__rq_seq) + 1) % (IPMI_LAN_REQUESTER_SEQUENCE_NUMBER_MAX + 1)))

#define IPMI_LAN_BACKOFF_COUNT 2

static void
_ipmi_lan_dump (ipmi_device_t dev,
		fiid_obj_t obj_rmcp,
		fiid_obj_t obj_session,
		fiid_obj_t obj_msg,
		fiid_obj_t obj_cmd,
		fiid_obj_t obj_trlr)
{
  char *rmcp_hdr =
    "RMCP Header:\n"
    "------------";
  char *session_hdr =
    "IPMI Session Header:\n"
    "--------------------";
  char *msg_hdr =
    "IPMI Message Header:\n"
    "--------------------";
  char *cmd_hdr =
    "IPMI Command Data:\n"
    "------------------";
  char *trlr_hdr =
    "IPMI Trailer:\n"
    "--------------";

  assert (dev
	  && dev->magic == IPMI_UDM_DEVICE_MAGIC
	  && (dev->flags & IPMI_FLAGS_DEBUG_DUMP)
	  && fiid_obj_valid(obj_rmcp)
	  && fiid_obj_valid(obj_session)
	  && fiid_obj_valid(obj_msg)
	  && fiid_obj_valid(obj_cmd));

  ipmi_obj_dump_perror(STDERR_FILENO,
		       NULL,
		       rmcp_hdr,
		       NULL,
		       obj_rmcp);
  ipmi_obj_dump_perror(STDERR_FILENO,
		       NULL,
		       session_hdr,
		       NULL,
		       obj_session);
  ipmi_obj_dump_perror(STDERR_FILENO,
		       NULL,
		       msg_hdr,
		       NULL,
		       obj_msg);
  ipmi_obj_dump_perror(STDERR_FILENO,
		       NULL,
		       cmd_hdr,
		       NULL,
		       obj_cmd);
  if (obj_trlr && fiid_obj_valid(obj_trlr))
    ipmi_obj_dump_perror(STDERR_FILENO,
			 NULL,
			 trlr_hdr,
			 NULL,
			 obj_trlr);
}

static void
_ipmi_lan_dump_rq (ipmi_device_t dev, fiid_obj_t obj_cmd_rq)
{
  assert (dev
	  && dev->magic == IPMI_UDM_DEVICE_MAGIC
	  && (dev->flags & IPMI_FLAGS_DEBUG_DUMP)
	  && fiid_obj_valid(obj_cmd_rq));

  _ipmi_lan_dump (dev,
		  dev->io.outofband.rq.obj_rmcp_hdr,
		  dev->io.outofband.rq.obj_lan_session_hdr,
		  dev->io.outofband.rq.obj_lan_msg_hdr,
		  obj_cmd_rq,
		  NULL);   
}

static void
_ipmi_lan_dump_rs (ipmi_device_t dev, fiid_obj_t obj_cmd_rs)
{
  assert (dev
	  && dev->magic == IPMI_UDM_DEVICE_MAGIC
	  && (dev->flags & IPMI_FLAGS_DEBUG_DUMP)
	  && fiid_obj_valid(obj_cmd_rs));

  _ipmi_lan_dump (dev,
		  dev->io.outofband.rs.obj_rmcp_hdr,
		  dev->io.outofband.rs.obj_lan_session_hdr,
		  dev->io.outofband.rs.obj_lan_msg_hdr,
		  obj_cmd_rs,
		  dev->io.outofband.rs.obj_lan_msg_trlr);   
}

static int8_t
_ipmi_lan_cmd_send (ipmi_device_t dev, fiid_obj_t obj_cmd_rq)
{
  uint8_t *pkt;
  int32_t pkt_len = 1024;
  int32_t send_len = 0;
  uint8_t authentication_type;

  assert(dev
	 && dev->magic == IPMI_UDM_DEVICE_MAGIC
	 && dev->io.outofband.local_sockfd 
	 && fiid_obj_valid(obj_cmd_rq)
	 && fiid_obj_packet_valid(obj_cmd_rq));

  FIID_OBJ_CLEAR(dev->io.outofband.rq.obj_rmcp_hdr);
  FIID_OBJ_CLEAR(dev->io.outofband.rq.obj_lan_session_hdr);
  FIID_OBJ_CLEAR(dev->io.outofband.rq.obj_lan_msg_hdr);
  FIID_OBJ_CLEAR(dev->io.outofband.rq.obj_lan_msg_trlr);   
    
  pkt = alloca (pkt_len);
  ERR (pkt);
  memset (pkt, 0, pkt_len);
    
  ERR (fill_rmcp_hdr_ipmi (dev->io.outofband.rq.obj_rmcp_hdr) != -1);
  ERR (fill_lan_msg_hdr (dev->net_fn,
			 dev->lun,
			 dev->io.outofband.rq_seq,
			 dev->io.outofband.rq.obj_lan_msg_hdr) != -1);

  if ((dev->io.outofband.lan_session_state == IPMI_LAN_SESSION_STATE_SET_SESSION_PRIVILEGE_LEVEL
       || dev->io.outofband.lan_session_state == IPMI_LAN_SESSION_STATE_IN_SESSION
       || dev->io.outofband.lan_session_state == IPMI_LAN_SESSION_STATE_CLOSE_SESSION)
      && dev->io.outofband.per_msg_auth_disabled)
    authentication_type = IPMI_AUTHENTICATION_TYPE_NONE;
  else
    authentication_type = dev->io.outofband.authentication_type;

  ERR (fill_lan_session_hdr (authentication_type,
			     dev->io.outofband.session_sequence_number,
			     dev->io.outofband.session_id,
			     dev->io.outofband.rq.obj_lan_session_hdr) != -1);
  ERR ((send_len = assemble_ipmi_lan_pkt (dev->io.outofband.rq.obj_rmcp_hdr,
					  dev->io.outofband.rq.obj_lan_session_hdr,
					  dev->io.outofband.rq.obj_lan_msg_hdr,
					  obj_cmd_rq,
					  dev->io.outofband.password,
					  IPMI_MAX_AUTHENTICATION_CODE_LENGTH,
					  pkt,
					  pkt_len)) != -1);

  if (dev->flags & IPMI_FLAGS_DEBUG_DUMP)
    _ipmi_lan_dump_rq (dev, obj_cmd_rq);

  ERR (!(ipmi_lan_sendto (dev->io.outofband.local_sockfd, 
			  pkt, 
			  send_len, 
			  0, 
			  (struct sockaddr *)&(dev->io.outofband.remote_host), 
			  sizeof(struct sockaddr_in)) < 0));

  ERR (!(gettimeofday(&dev->io.outofband.last_send, NULL) < 0));
  
  return (0);
}

static int
_session_timed_out(ipmi_device_t dev)
{
  struct timeval current;
  struct timeval session_timeout;
  struct timeval session_timeout_len;

  assert(dev && dev->magic == IPMI_UDM_DEVICE_MAGIC);

  session_timeout_len.tv_sec = dev->io.outofband.session_timeout / 1000;
  session_timeout_len.tv_usec = (dev->io.outofband.session_timeout - (session_timeout_len.tv_sec * 1000)) * 1000;
  timeradd(&dev->io.outofband.last_received, &session_timeout_len, &session_timeout);

  ERR (!(gettimeofday(&current, NULL) < 0));

  return timercmp(&current, &session_timeout, >);
}

static int
_calculate_timeout(ipmi_device_t dev, struct timeval *timeout)
{
  struct timeval current;
  struct timeval session_timeout;
  struct timeval session_timeout_len;
  struct timeval session_timeout_val;
  struct timeval retry_timeout;
  struct timeval retry_timeout_len;
  struct timeval retry_timeout_val;
  unsigned int retry_timeout_multiplier;

  assert(dev 
	 && dev->magic == IPMI_UDM_DEVICE_MAGIC
	 && timeout);

  ERR (!(gettimeofday(&current, NULL) < 0));

  session_timeout_len.tv_sec = dev->io.outofband.session_timeout / 1000;
  session_timeout_len.tv_usec = (dev->io.outofband.session_timeout - (session_timeout_len.tv_sec * 1000)) * 1000;

  timeradd(&current, &session_timeout_len, &session_timeout);
  timersub(&session_timeout, &current, &session_timeout_val);

  retry_timeout_multiplier = (dev->io.outofband.retry_count / IPMI_LAN_BACKOFF_COUNT) + 1;
  
  retry_timeout_len.tv_sec = (retry_timeout_multiplier * dev->io.outofband.retry_timeout) / 1000;
  retry_timeout_len.tv_usec = ((retry_timeout_multiplier * dev->io.outofband.retry_timeout) - (retry_timeout_len.tv_sec * 1000)) * 1000;

  timeradd(&dev->io.outofband.last_send, &retry_timeout_len, &retry_timeout);
  timersub(&retry_timeout, &current, &retry_timeout_val);

  if (timercmp(&retry_timeout_val, &session_timeout_val, <))
    {
      timeout->tv_sec = retry_timeout_val.tv_sec;
      timeout->tv_usec = retry_timeout_val.tv_usec;
    }
  else
    {
      timeout->tv_sec = session_timeout_val.tv_sec;
      timeout->tv_usec = session_timeout_val.tv_usec;
    }

  return 0;
}

static int8_t
_ipmi_lan_cmd_recv (ipmi_device_t dev, fiid_obj_t obj_cmd_rs)
{
  struct sockaddr_in from;
  socklen_t fromlen = 0;
  struct timeval timeout;
    
  uint8_t *pkt = NULL;
  uint32_t pkt_len = 1024;
  int32_t recv_len = 0;
  
  fd_set read_set;
  int status = 0;
  
  assert(dev
	 && dev->magic == IPMI_UDM_DEVICE_MAGIC
	 && dev->io.outofband.local_sockfd 
	 && fiid_obj_valid(obj_cmd_rs));
  
  pkt = alloca (pkt_len);
  ERR (pkt);
  
  FIID_OBJ_CLEAR(dev->io.outofband.rs.obj_rmcp_hdr); 
  FIID_OBJ_CLEAR(dev->io.outofband.rs.obj_lan_session_hdr); 
  FIID_OBJ_CLEAR(dev->io.outofband.rs.obj_lan_msg_hdr); 
  FIID_OBJ_CLEAR(dev->io.outofband.rs.obj_lan_msg_trlr); 
  
  memset (pkt, 0, pkt_len);
      
  if (dev->io.outofband.retry_timeout != 0)
    {
      FD_ZERO (&read_set);
      FD_SET (dev->io.outofband.local_sockfd, &read_set);
      
      ERR (!(_calculate_timeout(dev, &timeout) < 0));
      
      status = select ((dev->io.outofband.local_sockfd + 1), 
                       &read_set, NULL, NULL, &timeout);
          
      if (status == -1)
        return (-1);
	  
      if (status == 0)
        return (0); // resend the request
    }

  ERR (!((recv_len = ipmi_lan_recvfrom (dev->io.outofband.local_sockfd, 
                                        pkt, 
                                        pkt_len, 
                                        0, 
                                        (struct sockaddr *) &from, 
                                        &fromlen)) < 0));
  
  ERR (unassemble_ipmi_lan_pkt(pkt,
                               recv_len,
                               dev->io.outofband.rs.obj_rmcp_hdr,
                               dev->io.outofband.rs.obj_lan_session_hdr,
                               dev->io.outofband.rs.obj_lan_msg_hdr,
                               obj_cmd_rs,
                               dev->io.outofband.rs.obj_lan_msg_trlr) != -1);

  if (dev->flags & IPMI_FLAGS_DEBUG_DUMP)
    _ipmi_lan_dump_rs (dev, obj_cmd_rs);

  return (recv_len);
}

int8_t 
ipmi_lan_cmd (ipmi_device_t dev, 
	      fiid_obj_t obj_cmd_rq,
	      fiid_obj_t obj_cmd_rs)
{
  int status = 0;
  int retval = -1;
  int ret;

  ERR_EINVAL (dev 
	      && dev->magic == IPMI_UDM_DEVICE_MAGIC
	      && dev->io.outofband.local_sockfd 
	      && fiid_obj_valid(obj_cmd_rq)
	      && fiid_obj_valid(obj_cmd_rs));
  
  FIID_OBJ_PACKET_VALID(obj_cmd_rq);
  
  if (!dev->io.outofband.last_received.tv_sec
      && !dev->io.outofband.last_received.tv_usec)
    ERR (!(gettimeofday(&dev->io.outofband.last_received, NULL) < 0));

  dev->io.outofband.retry_count = 0;

  if (_ipmi_lan_cmd_send (dev, obj_cmd_rq) < 0)
    goto cleanup;

  while (1)
    {
      uint8_t authentication_type;

      if (_session_timed_out(dev))
        {
          errno = ETIMEDOUT;
          retval = -1;
          break;
        }
     
      if ((status = _ipmi_lan_cmd_recv (dev, obj_cmd_rs)) < 0)
        {
          retval = -1;
          break;
        }

      if (!status)
        {
          dev->io.outofband.session_sequence_number++;
          IPMI_LAN_RQ_SEQ_INC (dev->io.outofband.rq_seq);
          dev->io.outofband.retry_count++;

          if (_ipmi_lan_cmd_send (dev, obj_cmd_rq) < 0)
            goto cleanup;

          continue;
        }

      /* else received a packet */

      if ((dev->io.outofband.lan_session_state == IPMI_LAN_SESSION_STATE_SET_SESSION_PRIVILEGE_LEVEL
           || dev->io.outofband.lan_session_state == IPMI_LAN_SESSION_STATE_IN_SESSION
           || dev->io.outofband.lan_session_state == IPMI_LAN_SESSION_STATE_CLOSE_SESSION)
          && dev->io.outofband.per_msg_auth_disabled)
        authentication_type = IPMI_AUTHENTICATION_TYPE_NONE;
      else
        authentication_type = dev->io.outofband.authentication_type;
      
      ERR (!((ret = ipmi_lan_check_session_authentication_code (dev->io.outofband.rs.obj_lan_session_hdr,
                                                                dev->io.outofband.rs.obj_lan_msg_hdr,
                                                                obj_cmd_rs,
                                                                dev->io.outofband.rs.obj_lan_msg_trlr,
                                                                authentication_type,
                                                                dev->io.outofband.password,
                                                                IPMI_MAX_AUTHENTICATION_CODE_LENGTH)) < 0));

      if (!ret)
        continue;
  
      ERR (!((ret = ipmi_lan_check_rq_seq(dev->io.outofband.rs.obj_lan_msg_hdr, 
                                          dev->io.outofband.rq_seq)) < 0));

      if (!ret)
        continue;

      ERR (!((ret = ipmi_lan_check_checksum (dev->io.outofband.rs.obj_lan_msg_hdr,
                                             obj_cmd_rs,
                                             dev->io.outofband.rs.obj_lan_msg_trlr)) < 0));

      if (!ret)
        continue;
      
      ERR (!(gettimeofday(&dev->io.outofband.last_received, NULL) < 0));
      retval = status;
      break;
    }

 cleanup:
  dev->io.outofband.session_sequence_number++;
  IPMI_LAN_RQ_SEQ_INC (dev->io.outofband.rq_seq);
  return (retval);
}

int32_t 
ipmi_lan_cmd_raw (ipmi_device_t dev, 
		  uint8_t *buf_rq, 
		  size_t buf_rq_len, 
		  uint8_t *buf_rs, 
		  size_t buf_rs_len)
{
  fiid_obj_t obj_cmd_rq = NULL;
  fiid_obj_t obj_cmd_rs = NULL;
  int status = 0;
  int retval = -1;
  int32_t len;
  int ret;

  ERR_EINVAL (dev  
	      && dev->magic == IPMI_UDM_DEVICE_MAGIC
	      && dev->io.outofband.local_sockfd
	      && buf_rq
	      && buf_rq_len > 0
	      && buf_rs
	      && buf_rs_len > 0);

  FIID_OBJ_CREATE_CLEANUP(obj_cmd_rq, tmpl_lan_raw);
  FIID_OBJ_CREATE_CLEANUP(obj_cmd_rs, tmpl_lan_raw);
  FIID_OBJ_SET_ALL_CLEANUP (obj_cmd_rq, buf_rq, buf_rq_len);

  if (!dev->io.outofband.last_received.tv_sec
      && !dev->io.outofband.last_received.tv_usec)
    ERR (!(gettimeofday(&dev->io.outofband.last_received, NULL) < 0));

  dev->io.outofband.retry_count = 0;

  if (_ipmi_lan_cmd_send (dev, obj_cmd_rq) < 0)
    goto cleanup;

  while (1)
    {
      if (_session_timed_out(dev))
        {
          errno = ETIMEDOUT;
          retval = -1;
          break;
        }

      if ((status = _ipmi_lan_cmd_recv (dev, obj_cmd_rs)) < 0)
        {
          retval = -1;
          break;
        }

      if (!status)
        {
          dev->io.outofband.session_sequence_number++;
          IPMI_LAN_RQ_SEQ_INC (dev->io.outofband.rq_seq);
          dev->io.outofband.retry_count++;

          if (_ipmi_lan_cmd_send (dev, obj_cmd_rq) < 0)
            goto cleanup;

          continue;
        }

      /* else received a packet */

      ERR (!((ret = ipmi_lan_check_session_authentication_code (dev->io.outofband.rs.obj_lan_session_hdr,
                                                                dev->io.outofband.rs.obj_lan_msg_hdr,
                                                                obj_cmd_rs,
                                                                dev->io.outofband.rs.obj_lan_msg_trlr,
                                                                dev->io.outofband.authentication_type,
                                                                dev->io.outofband.password,
                                                                IPMI_MAX_AUTHENTICATION_CODE_LENGTH)) < 0));

      if (!ret)
        continue;
  
      ERR (!((ret = ipmi_lan_check_rq_seq(dev->io.outofband.rs.obj_lan_msg_hdr, 
                                          dev->io.outofband.rq_seq)) < 0));

      if (!ret)
        continue;

      ERR (!((ret = ipmi_lan_check_checksum (dev->io.outofband.rs.obj_lan_msg_hdr,
                                             obj_cmd_rs,
                                             dev->io.outofband.rs.obj_lan_msg_trlr)) < 0));

      if (!ret)
        continue;
      
      ERR (!(gettimeofday(&dev->io.outofband.last_received, NULL) < 0));

      FIID_OBJ_GET_ALL_LEN_CLEANUP (len, obj_cmd_rs, buf_rs, buf_rs_len);
      retval = len;
      break;
    }

  dev->io.outofband.session_sequence_number++;
  IPMI_LAN_RQ_SEQ_INC (dev->io.outofband.rq_seq);
 cleanup:
  FIID_OBJ_DESTROY_NO_RETURN (obj_cmd_rq);
  FIID_OBJ_DESTROY_NO_RETURN (obj_cmd_rs);
  return (retval);
}
