/* 
   ipmi-messaging-support-cmds-udm.c - IPMI UDM Message Support Commands

   Copyright (C) 2003, 2004, 2005 FreeIPMI Core Team

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#ifdef STDC_HEADERS
#include <string.h>
#endif /* STDC_HEADERS */
#include <errno.h>
#if TIME_WITH_SYS_TIME
#include <sys/time.h>
#include <time.h>
#else /* !TIME_WITH_SYS_TIME */
#if HAVE_SYS_TIME_H
#include <sys/time.h>
#else /* !HAVE_SYS_TIME_H */
#include <time.h>
#endif /* !HAVE_SYS_TIME_H */
#endif	/* !TIME_WITH_SYS_TIME */

#include "freeipmi/udm/ipmi-messaging-support-cmds-udm.h"
#include "freeipmi/ipmi-messaging-support-cmds.h"
#include "freeipmi/ipmi-authentication-type-spec.h"
#include "freeipmi/ipmi-channel-spec.h"
#include "freeipmi/ipmi-comp-code-spec.h"
#include "freeipmi/ipmi-device-global-cmds.h"
#include "freeipmi/ipmi-ipmb-interface.h"
#include "freeipmi/ipmi-netfn-spec.h"
#include "freeipmi/ipmi-privilege-level-spec.h"
#include "freeipmi/udm/ipmi-device-global-cmds-udm.h"

#include "ipmi-udm-device.h"

#include "err-wrappers.h"
#include "fiid-wrappers.h"
#include "freeipmi-portability.h"

int8_t 
ipmi_cmd_get_channel_authentication_capabilities (ipmi_device_t dev, 
						  fiid_obj_t obj_cmd_rs)
{
  struct ipmi_device local_dev;
  fiid_obj_t obj_cmd_rq = NULL;
  int8_t rv = -1;

  ERR_EINVAL (dev 
	      && dev->magic == IPMI_UDM_DEVICE_MAGIC
	      && dev->type == IPMI_DEVICE_LAN
	      && fiid_obj_valid(obj_cmd_rs));

  FIID_OBJ_TEMPLATE_COMPARE(obj_cmd_rs, tmpl_cmd_get_channel_authentication_capabilities_rs);
  
  memcpy(&local_dev, dev, sizeof(struct ipmi_device));
  local_dev.io.outofband.authentication_type = IPMI_AUTHENTICATION_TYPE_NONE;

  FIID_OBJ_CREATE (obj_cmd_rq, tmpl_cmd_get_channel_authentication_capabilities_rq);

  ERR_CLEANUP (!(fill_cmd_get_channel_authentication_capabilities (IPMI_CHANNEL_NUMBER_CURRENT_CHANNEL,
								   IPMI_PRIVILEGE_LEVEL_USER, 
								   obj_cmd_rq) < 0));

  dev->lun = IPMI_BMC_IPMB_LUN_BMC;
  dev->net_fn = IPMI_NET_FN_APP_RQ;
  ERR_IPMI_CMD_CLEANUP (&local_dev, 
			IPMI_BMC_IPMB_LUN_BMC, 
			IPMI_NET_FN_APP_RQ, 
			obj_cmd_rq, 
			obj_cmd_rs);  
                       
  rv = 0;
 cleanup:
  FIID_OBJ_DESTROY_NO_RETURN(obj_cmd_rq);
  return (rv);
}

int8_t 
ipmi_cmd_get_session_challenge (ipmi_device_t dev, 
				fiid_obj_t obj_cmd_rs)
{
  struct ipmi_device local_dev;
  fiid_obj_t obj_cmd_rq = NULL;
  int8_t rv = -1;

  ERR_EINVAL (dev 
	      && dev->magic == IPMI_UDM_DEVICE_MAGIC
	      && dev->type == IPMI_DEVICE_LAN
	      && fiid_obj_valid(obj_cmd_rs));

  FIID_OBJ_TEMPLATE_COMPARE(obj_cmd_rs, tmpl_cmd_get_session_challenge_rs);
  
  memcpy(&local_dev, dev, sizeof(struct ipmi_device));
  local_dev.io.outofband.authentication_type = IPMI_AUTHENTICATION_TYPE_NONE;

  FIID_OBJ_CREATE (obj_cmd_rq, tmpl_cmd_get_session_challenge_rq);

  ERR_CLEANUP (!(fill_cmd_get_session_challenge (dev->io.outofband.authentication_type, 
						 dev->io.outofband.username,
						 IPMI_MAX_USER_NAME_LENGTH,
						 obj_cmd_rq) < 0));

  dev->lun = IPMI_BMC_IPMB_LUN_BMC;
  dev->net_fn = IPMI_NET_FN_APP_RQ;
  ERR_CLEANUP (!(ipmi_cmd (&local_dev, 
                           IPMI_BMC_IPMB_LUN_BMC, 
                           IPMI_NET_FN_APP_RQ, 
                           obj_cmd_rq, 
                           obj_cmd_rs) < 0));
  if (ipmi_check_completion_code_success(obj_cmd_rs) != 1)
    {
      errno = EACCES;
      goto cleanup;
    }
  
  rv = 0;
 cleanup:
  FIID_OBJ_DESTROY_NO_RETURN(obj_cmd_rq);
  return (rv);
}

int8_t 
ipmi_cmd_activate_session (ipmi_device_t dev, 
			   fiid_obj_t obj_cmd_rs)
{
  uint32_t initial_outbound_sequence_number = 0;
  fiid_obj_t obj_cmd_rq = NULL;
  int8_t rv = -1;

  ERR_EINVAL (dev 
	      && dev->magic == IPMI_UDM_DEVICE_MAGIC
	      && dev->type == IPMI_DEVICE_LAN
	      && fiid_obj_valid(obj_cmd_rs));

  FIID_OBJ_TEMPLATE_COMPARE(obj_cmd_rs, tmpl_cmd_activate_session_rs);
  
  { 
    /* Random number generation */
    unsigned int seedp;
    seedp = (unsigned int) clock () + (unsigned int) time (NULL);
    srand (seedp);
    initial_outbound_sequence_number = rand_r (&seedp);
  }

  FIID_OBJ_CREATE (obj_cmd_rq, tmpl_cmd_activate_session_rq);

  ERR_CLEANUP (!(fill_cmd_activate_session (dev->io.outofband.authentication_type, 
					    dev->io.outofband.privilege_level, 
					    dev->io.outofband.challenge_string, 
					    IPMI_CHALLENGE_STRING_LENGTH, 
					    initial_outbound_sequence_number, 
					    obj_cmd_rq) < 0));

  dev->lun = IPMI_BMC_IPMB_LUN_BMC;
  dev->net_fn = IPMI_NET_FN_APP_RQ;
  ERR_CLEANUP (!(ipmi_cmd (dev, 
                           IPMI_BMC_IPMB_LUN_BMC, 
                           IPMI_NET_FN_APP_RQ, 
                           obj_cmd_rq, 
                           obj_cmd_rs) < 0));
  if (ipmi_check_completion_code_success (obj_cmd_rs) != 1)
    {
      if (ipmi_check_completion_code(obj_cmd_rs, IPMI_COMP_CODE_NO_SESSION_SLOT_AVAILABLE) == 1
          || ipmi_check_completion_code(obj_cmd_rs, IPMI_COMP_CODE_NO_SLOT_AVAILABLE_FOR_GIVEN_USER) == 1
          || ipmi_check_completion_code(obj_cmd_rs, IPMI_COMP_CODE_NO_SLOT_AVAILABLE_TO_SUPPORT_USER) == 1)
        errno = EBUSY;
      else if (ipmi_check_completion_code(obj_cmd_rs, IPMI_COMP_CODE_EXCEEDS_PRIVILEGE_LEVEL) == 1)
        errno = EACCES;
      else
        errno = EIO;
      goto cleanup;
    }

  rv = 0;
 cleanup:
  FIID_OBJ_DESTROY_NO_RETURN(obj_cmd_rq);
  return (rv);
}

int8_t 
ipmi_cmd_set_session_privilege_level (ipmi_device_t dev, 
				      fiid_obj_t obj_cmd_rs)
{
  fiid_obj_t obj_cmd_rq = NULL;
  int8_t rv = -1;

  ERR_EINVAL (dev 
	      && dev->magic == IPMI_UDM_DEVICE_MAGIC
	      && dev->type == IPMI_DEVICE_LAN
	      && fiid_obj_valid(obj_cmd_rs));

  FIID_OBJ_TEMPLATE_COMPARE(obj_cmd_rs, tmpl_cmd_set_session_privilege_level_rs);
  
  FIID_OBJ_CREATE (obj_cmd_rq, tmpl_cmd_set_session_privilege_level_rq);

  ERR_CLEANUP (!(fill_cmd_set_session_privilege_level (dev->io.outofband.privilege_level, 
						       obj_cmd_rq) < 0));

  dev->lun = IPMI_BMC_IPMB_LUN_BMC;
  dev->net_fn = IPMI_NET_FN_APP_RQ;
  ERR_CLEANUP (!(ipmi_cmd (dev, 
                           IPMI_BMC_IPMB_LUN_BMC, 
                           IPMI_NET_FN_APP_RQ, 
                           obj_cmd_rq, 
                           obj_cmd_rs) < 0));
  if (ipmi_check_completion_code_success (obj_cmd_rs) != 1)
    {
      if (ipmi_check_completion_code(obj_cmd_rs, IPMI_COMP_CODE_RQ_LEVEL_NOT_AVAILABLE_FOR_USER) == 1
          || ipmi_check_completion_code(obj_cmd_rs, IPMI_COMP_CODE_RQ_LEVEL_EXCEEDS_USER_PRIVILEGE_LIMIT) == 1)
        errno = EACCES;
      else
        errno = EIO;
      goto cleanup;
    }
  
  rv = 0;
 cleanup:
  FIID_OBJ_DESTROY_NO_RETURN(obj_cmd_rq);
  return (rv);
}

int8_t 
ipmi_lan_close_session (ipmi_device_t dev, 
			fiid_obj_t obj_cmd_rs)
{
  fiid_obj_t obj_cmd_rq = NULL;
  int8_t rv = -1;

  ERR_EINVAL (dev
	      && dev->magic == IPMI_UDM_DEVICE_MAGIC
	      && dev->type == IPMI_DEVICE_LAN
	      && fiid_obj_valid(obj_cmd_rs));
  
  FIID_OBJ_TEMPLATE_COMPARE(obj_cmd_rs, tmpl_cmd_close_session_rs);
  
  FIID_OBJ_CREATE (obj_cmd_rq, tmpl_cmd_close_session_rq);

  ERR_CLEANUP (!(fill_cmd_close_session (dev->io.outofband.session_id, obj_cmd_rq) < 0));

  ERR_IPMI_CMD_CLEANUP (dev,
			IPMI_BMC_IPMB_LUN_BMC,
			IPMI_NET_FN_APP_RQ,
			obj_cmd_rq,
			obj_cmd_rs);

  rv = 0;
 cleanup:
  FIID_OBJ_DESTROY_NO_RETURN(obj_cmd_rq);
  return (rv);
}

int8_t 
ipmi_cmd_set_channel_access (ipmi_device_t dev, 
			     uint8_t channel_number, 
			     uint8_t ipmi_messaging_access_mode, 
			     uint8_t user_level_authentication, 
			     uint8_t per_message_authentication, 
			     uint8_t pef_alerting, 
			     uint8_t channel_access_set, 
			     uint8_t channel_privilege_level_limit, 
			     uint8_t channel_privilege_level_limit_set, 
			     fiid_obj_t obj_cmd_rs)
{
  fiid_obj_t obj_cmd_rq = NULL;
  int8_t rv = -1;

  ERR_EINVAL (dev
	      && dev->magic == IPMI_UDM_DEVICE_MAGIC
	      && IPMI_CHANNEL_NUMBER_VALID(channel_number)
	      && IPMI_MESSAGING_ACCESS_MODE_VALID(ipmi_messaging_access_mode)
	      && IPMI_USER_LEVEL_AUTHENTICATION_VALID(user_level_authentication)
	      && IPMI_PER_MESSAGE_AUTHENTICATION_VALID(per_message_authentication)
	      && IPMI_PEF_ALERTING_VALID(pef_alerting)
	      && IPMI_CHANNEL_ACCESS_VALID(channel_access_set)
	      && IPMI_PRIVILEGE_LEVEL_LIMIT_VALID(channel_privilege_level_limit)
	      && IPMI_PRIVILEGE_LEVEL_LIMIT_SET_VALID(channel_privilege_level_limit_set)
	      && fiid_obj_valid(obj_cmd_rs));
  
  FIID_OBJ_TEMPLATE_COMPARE(obj_cmd_rs, tmpl_cmd_set_channel_access_rs);
  
  FIID_OBJ_CREATE (obj_cmd_rq, tmpl_cmd_set_channel_access_rq);

  ERR_CLEANUP (!(fill_cmd_set_channel_access (channel_number, 
					      ipmi_messaging_access_mode, 
					      user_level_authentication, 
					      per_message_authentication, 
					      pef_alerting, 
					      channel_access_set, 
					      channel_privilege_level_limit, 
					      channel_privilege_level_limit_set,
					      obj_cmd_rq) < 0));

  ERR_IPMI_CMD_CLEANUP (dev, 
			IPMI_BMC_IPMB_LUN_BMC, 
			IPMI_NET_FN_APP_RQ, 
			obj_cmd_rq, 
			obj_cmd_rs);

  rv = 0;
 cleanup:
  FIID_OBJ_DESTROY_NO_RETURN(obj_cmd_rq);
  return (rv);
}

int8_t 
ipmi_cmd_get_channel_access (ipmi_device_t dev, 
			     uint8_t channel_number,
			     uint8_t channel_access_get,
			     fiid_obj_t obj_cmd_rs)
{
  fiid_obj_t obj_cmd_rq = NULL;
  int8_t rv = -1;
  
  ERR_EINVAL (dev
	      && dev->magic == IPMI_UDM_DEVICE_MAGIC
	      && IPMI_CHANNEL_NUMBER_VALID(channel_number)
	      && IPMI_CHANNEL_ACCESS_GET_VALID(channel_access_get)
	      && fiid_obj_valid(obj_cmd_rs));
  
  FIID_OBJ_TEMPLATE_COMPARE(obj_cmd_rs, tmpl_cmd_get_channel_access_rs);
  
  FIID_OBJ_CREATE (obj_cmd_rq, tmpl_cmd_get_channel_access_rq);

  ERR_CLEANUP (!(fill_cmd_get_channel_access (channel_number, 
					      channel_access_get,
					      obj_cmd_rq) < 0));

  ERR_IPMI_CMD_CLEANUP (dev, 
			IPMI_BMC_IPMB_LUN_BMC, 
			IPMI_NET_FN_APP_RQ, 
			obj_cmd_rq, 
			obj_cmd_rs);

  rv = 0;
 cleanup:
  FIID_OBJ_DESTROY_NO_RETURN(obj_cmd_rq);
  return (rv);
}

int8_t 
ipmi_cmd_get_channel_info (ipmi_device_t dev, 
			   uint8_t channel_number,
			   fiid_obj_t obj_cmd_rs)
{
  fiid_obj_t obj_cmd_rq = NULL;
  int8_t rv = -1;

  ERR_EINVAL (dev
	      && dev->magic == IPMI_UDM_DEVICE_MAGIC
	      && IPMI_CHANNEL_NUMBER_VALID(channel_number)
	      && fiid_obj_valid(obj_cmd_rs));

  FIID_OBJ_TEMPLATE_COMPARE(obj_cmd_rs, tmpl_cmd_get_channel_info_rs);
  
  FIID_OBJ_CREATE (obj_cmd_rq, tmpl_cmd_get_channel_info_rq);

  ERR_CLEANUP (!(fill_cmd_get_channel_info (channel_number, obj_cmd_rq) < 0));

  ERR_IPMI_CMD_CLEANUP (dev,
			IPMI_BMC_IPMB_LUN_BMC,
			IPMI_NET_FN_APP_RQ,
			obj_cmd_rq,
			obj_cmd_rs);

  rv = 0;
 cleanup:
  FIID_OBJ_DESTROY_NO_RETURN(obj_cmd_rq);
  return (rv);
}

int8_t
ipmi_cmd_set_channel_security_keys (ipmi_device_t dev,
                                    uint8_t channel_number,
                                    uint8_t operation,
                                    uint8_t key_id,
                                    uint8_t *key_value,
                                    uint32_t key_value_len,
                                    fiid_obj_t obj_cmd_rs)
{
  fiid_obj_t obj_cmd_rq = NULL;
  int8_t rv = -1;
  
  ERR_EINVAL (dev
	      && dev->magic == IPMI_UDM_DEVICE_MAGIC
	      && IPMI_CHANNEL_NUMBER_VALID(channel_number)
              && IPMI_CHANNEL_SECURITY_KEYS_OPERATION_VALID(operation)
              && IPMI_CHANNEL_SECURITY_KEYS_KEY_ID_VALID(key_id)
	      && !((key_id == IPMI_CHANNEL_SECURITY_KEYS_KEY_ID_K_R
		    && key_value)
		   && key_value_len > IPMI_MAX_K_R_LENGTH)
	      && !((key_id == IPMI_CHANNEL_SECURITY_KEYS_KEY_ID_K_G
		    && key_value)
		   && key_value_len > IPMI_MAX_K_G_LENGTH)
	      && fiid_obj_valid(obj_cmd_rs));
  
  FIID_OBJ_TEMPLATE_COMPARE(obj_cmd_rs, tmpl_cmd_set_channel_security_keys_rs);
  
  FIID_OBJ_CREATE (obj_cmd_rq, tmpl_cmd_set_channel_security_keys_rq);

  ERR_CLEANUP (!(fill_cmd_set_channel_security_keys (channel_number,
                                                     operation,
                                                     key_id,
                                                     key_value,
                                                     key_value_len,
                                                     obj_cmd_rq) < 0));
  
  ERR_IPMI_CMD_CLEANUP (dev, 
			IPMI_BMC_IPMB_LUN_BMC, 
			IPMI_NET_FN_APP_RQ, 
			obj_cmd_rq, 
			obj_cmd_rs);

  rv = 0;
 cleanup:
  FIID_OBJ_DESTROY_NO_RETURN(obj_cmd_rq);
  return (rv);
}

int8_t 
ipmi_cmd_set_user_access (ipmi_device_t dev, 
			  uint8_t channel_number,
			  uint8_t user_ipmi_messaging,
			  uint8_t user_link_authentication,
			  uint8_t user_restricted_to_callback,
			  uint8_t user_id,
			  uint8_t user_privilege_level_limit,
			  uint8_t user_session_number_limit, 
			  fiid_obj_t obj_cmd_rs)
{
  fiid_obj_t obj_cmd_rq = NULL;
  int8_t rv = -1;
  
  ERR_EINVAL (dev
	      && dev->magic == IPMI_UDM_DEVICE_MAGIC
	      && IPMI_CHANNEL_NUMBER_VALID(channel_number)
	      && IPMI_USER_IPMI_MESSAGING_VALID(user_ipmi_messaging)
	      && IPMI_USER_LINK_AUTHENTICATION_VALID(user_link_authentication)
	      && IPMI_USER_RESTRICTED_TO_CALLBACK_VALID(user_restricted_to_callback)
	      && IPMI_PRIVILEGE_LEVEL_LIMIT_VALID(user_privilege_level_limit)
	      && fiid_obj_valid(obj_cmd_rs));
  
  FIID_OBJ_TEMPLATE_COMPARE(obj_cmd_rs, tmpl_cmd_set_user_access_rs);
  
  FIID_OBJ_CREATE (obj_cmd_rq, tmpl_cmd_set_user_access_rq);

  ERR_CLEANUP (!(fill_cmd_set_user_access (channel_number,
					   user_ipmi_messaging,
					   user_link_authentication,
					   user_restricted_to_callback,
					   user_id,
					   user_privilege_level_limit,
					   user_session_number_limit,
					   obj_cmd_rq) < 0));

  ERR_IPMI_CMD_CLEANUP (dev, 
			IPMI_BMC_IPMB_LUN_BMC, 
			IPMI_NET_FN_APP_RQ, 
			obj_cmd_rq, 
			obj_cmd_rs);

  rv = 0;
 cleanup:
  FIID_OBJ_DESTROY_NO_RETURN(obj_cmd_rq);
  return (rv);
}

int8_t 
ipmi_cmd_get_user_access (ipmi_device_t dev, 
			  uint8_t channel_number,
			  uint8_t user_id,
			  fiid_obj_t obj_cmd_rs)
{
  fiid_obj_t obj_cmd_rq = NULL;
  int8_t rv = -1;
  
  ERR_EINVAL (dev
	      && dev->magic == IPMI_UDM_DEVICE_MAGIC
	      && IPMI_CHANNEL_NUMBER_VALID(channel_number)
	      && fiid_obj_valid(obj_cmd_rs));
  
  FIID_OBJ_TEMPLATE_COMPARE(obj_cmd_rs, tmpl_cmd_get_user_access_rs);
  
  FIID_OBJ_CREATE (obj_cmd_rq, tmpl_cmd_get_user_access_rq);

  ERR_CLEANUP (!(fill_cmd_get_user_access (channel_number, user_id, obj_cmd_rq) < 0));

  ERR_IPMI_CMD_CLEANUP (dev, 
			IPMI_BMC_IPMB_LUN_BMC, 
			IPMI_NET_FN_APP_RQ, 
			obj_cmd_rq, 
			obj_cmd_rs);

  rv = 0;
 cleanup:
  FIID_OBJ_DESTROY_NO_RETURN(obj_cmd_rq);
  return (rv);
}

int8_t 
ipmi_cmd_set_user_name (ipmi_device_t dev, 
			uint8_t user_id, 
			uint8_t *user_name, 
			unsigned int user_name_len,
			fiid_obj_t obj_cmd_rs)
{
  fiid_obj_t obj_cmd_rq = NULL;
  int8_t rv = -1;

  ERR_EINVAL (dev
	      && dev->magic == IPMI_UDM_DEVICE_MAGIC
	      && !(user_name && user_name_len > IPMI_MAX_USER_NAME_LENGTH)
	      && fiid_obj_valid(obj_cmd_rs));
  
  FIID_OBJ_TEMPLATE_COMPARE(obj_cmd_rs, tmpl_cmd_set_user_name_rs);
  
  FIID_OBJ_CREATE (obj_cmd_rq, tmpl_cmd_set_user_name_rq);

  ERR_CLEANUP (!(fill_cmd_set_user_name (user_id, 
					 user_name, 
					 user_name_len,
					 obj_cmd_rq) < 0));

  ERR_IPMI_CMD_CLEANUP (dev, 
			IPMI_BMC_IPMB_LUN_BMC, 
			IPMI_NET_FN_APP_RQ, 
			obj_cmd_rq, 
			obj_cmd_rs);

  rv = 0;
 cleanup:
  FIID_OBJ_DESTROY_NO_RETURN(obj_cmd_rq);
  return (rv);
}

int8_t 
ipmi_cmd_get_user_name (ipmi_device_t dev, 
			uint8_t user_id, 
			fiid_obj_t obj_cmd_rs)
{
  fiid_obj_t obj_cmd_rq = NULL;
  int8_t rv = -1;
  
  ERR_EINVAL (dev
	      && dev->magic == IPMI_UDM_DEVICE_MAGIC
	      && fiid_obj_valid(obj_cmd_rs));
  
  FIID_OBJ_TEMPLATE_COMPARE(obj_cmd_rs, tmpl_cmd_get_user_name_rs);
  
  FIID_OBJ_CREATE (obj_cmd_rq, tmpl_cmd_get_user_name_rq);

  ERR_CLEANUP (!(fill_cmd_get_user_name (user_id, obj_cmd_rq) < 0));

  ERR_IPMI_CMD_CLEANUP (dev, 
			IPMI_BMC_IPMB_LUN_BMC, 
			IPMI_NET_FN_APP_RQ, 
			obj_cmd_rq, 
			obj_cmd_rs);

  rv = 0;
 cleanup:
  FIID_OBJ_DESTROY_NO_RETURN(obj_cmd_rq);
  return (rv);
}

int8_t 
ipmi_cmd_set_user_password (ipmi_device_t dev, 
			    uint8_t user_id, 
			    uint8_t operation, 
			    char *password,
			    unsigned int password_len,
			    fiid_obj_t obj_cmd_rs)
{
  fiid_obj_t obj_cmd_rq = NULL;
  int8_t rv = -1;
  
  ERR_EINVAL (dev
	      && dev->magic == IPMI_UDM_DEVICE_MAGIC
	      && IPMI_PASSWORD_OPERATION_VALID(operation)
	      && !(password && password_len > IPMI_1_5_MAX_PASSWORD_LENGTH)
	      && fiid_obj_valid(obj_cmd_rs));
  
  FIID_OBJ_TEMPLATE_COMPARE(obj_cmd_rs, tmpl_cmd_set_user_password_rs);
  
  FIID_OBJ_CREATE (obj_cmd_rq, tmpl_cmd_set_user_password_rq);

  ERR_CLEANUP (!(fill_cmd_set_user_password (user_id, 
					     operation, 
					     password, 
					     password_len,
					     obj_cmd_rq) < 0));

  ERR_IPMI_CMD_CLEANUP (dev, 
			IPMI_BMC_IPMB_LUN_BMC, 
			IPMI_NET_FN_APP_RQ, 
			obj_cmd_rq, 
			obj_cmd_rs);
  
  rv = 0;
 cleanup:
  FIID_OBJ_DESTROY_NO_RETURN(obj_cmd_rq);
  return (rv);
}

int8_t 
ipmi_cmd_set_user_password_v20 (ipmi_device_t dev, 
                                uint8_t user_id, 
                                uint8_t password_size,
                                uint8_t operation, 
                                char *password,
                                unsigned int password_len,
                                fiid_obj_t obj_cmd_rs)
{
  fiid_obj_t obj_cmd_rq = NULL;
  int8_t rv = -1;
  
  ERR_EINVAL (dev
	      && dev->magic == IPMI_UDM_DEVICE_MAGIC
	      && IPMI_PASSWORD_OPERATION_VALID(operation)
              && IPMI_PASSWORD_SIZE_VALID(password_size)
              && !(password
                   && password_size == IPMI_PASSWORD_SIZE_16_BYTES
                   && password_len > IPMI_1_5_MAX_PASSWORD_LENGTH)
              && !(password
                   && password_size == IPMI_PASSWORD_SIZE_20_BYTES
                   && password_len > IPMI_2_0_MAX_PASSWORD_LENGTH)
	      && fiid_obj_valid(obj_cmd_rs));
  
  FIID_OBJ_TEMPLATE_COMPARE(obj_cmd_rs, tmpl_cmd_set_user_password_rs);
  
  FIID_OBJ_CREATE (obj_cmd_rq, tmpl_cmd_set_user_password_v20_rq);
  
  ERR_CLEANUP (!(fill_cmd_set_user_password_v20 (user_id, 
                                                 password_size,
                                                 operation, 
                                                 password, 
                                                 password_len,
                                                 obj_cmd_rq) < 0));

  ERR_IPMI_CMD_CLEANUP (dev, 
			IPMI_BMC_IPMB_LUN_BMC, 
			IPMI_NET_FN_APP_RQ, 
			obj_cmd_rq, 
			obj_cmd_rs);
  
  rv = 0;
 cleanup:
  FIID_OBJ_DESTROY_NO_RETURN(obj_cmd_rq);
  return (rv);
}

int8_t
ipmi_get_channel_number (ipmi_device_t dev, uint8_t channel_medium_type)
{
  fiid_obj_t obj_data_rs = NULL;
  uint64_t manufacturer_id, product_id;
  int8_t rv = -1;
  uint64_t val;
  int i;
    
  /* XXX channel medium type check? */
  ERR_EINVAL (dev && dev->magic == IPMI_UDM_DEVICE_MAGIC);

  if (channel_medium_type == IPMI_CHANNEL_MEDIUM_TYPE_LAN_802_3)
    {
      FIID_OBJ_CREATE_CLEANUP(obj_data_rs, tmpl_cmd_get_device_id_rs);
      
      ERR_CLEANUP (!(ipmi_cmd_get_device_id (dev, obj_data_rs) < 0));
      
      FIID_OBJ_GET_CLEANUP (obj_data_rs, "manufacturer_id.id", &manufacturer_id);

      FIID_OBJ_GET_CLEANUP (obj_data_rs, "product_id", &product_id);
      
      switch (manufacturer_id)
	{
	case IPMI_MANUFACTURER_ID_INTEL:
	case 0xB000157: // Intel 
	  switch (product_id)
	    {
	    case IPMI_PRODUCT_ID_SE7501WV2:
	      rv = 7;
	      goto cleanup;
	    }
	}

      FIID_OBJ_DESTROY_NO_RETURN(obj_data_rs);
    }
  
  FIID_OBJ_CREATE_CLEANUP(obj_data_rs, tmpl_cmd_get_channel_info_rs);
  
  /* Channel numbers range from 0 - 7 */
  for (i = 0; i < 8; i++)
    {
      if (ipmi_cmd_get_channel_info (dev, i, obj_data_rs) != 0)
	continue;
	
      FIID_OBJ_GET_CLEANUP (obj_data_rs, "channel_medium_type", &val);
      
      if ((uint8_t) val == channel_medium_type)
	{
	  FIID_OBJ_GET_CLEANUP (obj_data_rs, "actual_channel_number", &val);
	  
	  rv = (int8_t) val;
	  break;
	}
    }

 cleanup:
  FIID_OBJ_DESTROY_NO_RETURN(obj_data_rs);
  return (rv);
}

int8_t 
ipmi_lan_open_session (ipmi_device_t dev)
{
  fiid_obj_t obj_cmd_rs = NULL;
  uint64_t supported_authentication_type = 0;
  uint64_t temp_session_id = 0;
  uint8_t challenge_string[IPMI_CHALLENGE_STRING_LENGTH];
  uint64_t temp_session_sequence_number = 0;
  int8_t rv = -1;
  uint64_t val;

  ERR_EINVAL (dev 
	      && dev->magic == IPMI_UDM_DEVICE_MAGIC
	      && IPMI_AUTHENTICATION_TYPE_VALID(dev->io.outofband.authentication_type));
  
  dev->io.outofband.rq_seq = 0;
  
  FIID_OBJ_CREATE_CLEANUP(obj_cmd_rs, tmpl_cmd_get_channel_authentication_capabilities_rs);
  
  dev->io.outofband.lan_session_state = IPMI_LAN_SESSION_STATE_GET_CHANNEL_AUTHENTICAPTION_CAPABILITIES;
  ERR_CLEANUP (!(ipmi_cmd_get_channel_authentication_capabilities (dev, obj_cmd_rs) < 0));
  FIID_OBJ_GET_CLEANUP (obj_cmd_rs,
                        "authentication_status.per_message_authentication",
                        &val);
  dev->io.outofband.per_msg_auth_disabled = val;

  switch (dev->io.outofband.authentication_type)
    {
    case IPMI_AUTHENTICATION_TYPE_NONE:
      FIID_OBJ_GET_CLEANUP (obj_cmd_rs, 
			    "authentication_type.none", 
			    &supported_authentication_type);
      break;
    case IPMI_AUTHENTICATION_TYPE_MD2:
      FIID_OBJ_GET_CLEANUP (obj_cmd_rs, 
			    "authentication_type.md2", 
			    &supported_authentication_type);
      break;
    case IPMI_AUTHENTICATION_TYPE_MD5:
      FIID_OBJ_GET_CLEANUP (obj_cmd_rs, 
			    "authentication_type.md5", 
			    &supported_authentication_type);
      break;
    case IPMI_AUTHENTICATION_TYPE_STRAIGHT_PASSWORD_KEY:
      FIID_OBJ_GET_CLEANUP (obj_cmd_rs, 
			    "authentication_type.straight_password_key", 
			    &supported_authentication_type);
      break;
    case IPMI_AUTHENTICATION_TYPE_OEM_PROP:
      FIID_OBJ_GET_CLEANUP (obj_cmd_rs, 
			    "authentication_type.oem_prop", 
			    &supported_authentication_type);
      break;
    }

  ERR_ENOTSUP_CLEANUP (supported_authentication_type != 0);
  
  FIID_OBJ_DESTROY(obj_cmd_rs);
  FIID_OBJ_CREATE_CLEANUP (obj_cmd_rs, tmpl_cmd_get_session_challenge_rs);
  dev->io.outofband.lan_session_state = IPMI_LAN_SESSION_STATE_GET_SESSION_CHALLENGE;
  ERR_CLEANUP (!(ipmi_cmd_get_session_challenge (dev, obj_cmd_rs) < 0));

  FIID_OBJ_GET_CLEANUP (obj_cmd_rs, 
			"temp_session_id", 
			&temp_session_id);

  dev->io.outofband.session_id = temp_session_id;
  FIID_OBJ_GET_DATA_CLEANUP (obj_cmd_rs, 
			     "challenge_string", 
			     challenge_string,
			     IPMI_CHALLENGE_STRING_LENGTH);

  memcpy (dev->io.outofband.challenge_string, 
	  challenge_string, 
	  IPMI_CHALLENGE_STRING_LENGTH);

  FIID_OBJ_DESTROY(obj_cmd_rs);
  FIID_OBJ_CREATE_CLEANUP (obj_cmd_rs, tmpl_cmd_activate_session_rs);
  dev->io.outofband.lan_session_state = IPMI_LAN_SESSION_STATE_ACTIVATE_SESSION;
  ERR_CLEANUP (!(ipmi_cmd_activate_session (dev, obj_cmd_rs) < 0));

  FIID_OBJ_GET_CLEANUP (obj_cmd_rs, 
			"session_id", 
			&temp_session_id);

  dev->io.outofband.session_id = temp_session_id;
  FIID_OBJ_GET_CLEANUP (obj_cmd_rs, 
			"initial_inbound_sequence_number", 
			&temp_session_sequence_number);
  dev->io.outofband.session_sequence_number = temp_session_sequence_number;
  
  FIID_OBJ_GET_CLEANUP (obj_cmd_rs,
                        "authentication_type",
                        &val);

  /* achu: Make sure the BMC actually wants to do permsg auth */
  if (dev->io.outofband.per_msg_auth_disabled
      && val != IPMI_AUTHENTICATION_TYPE_NONE)
    dev->io.outofband.per_msg_auth_disabled = 0;

  FIID_OBJ_DESTROY(obj_cmd_rs);
  FIID_OBJ_CREATE_CLEANUP (obj_cmd_rs, tmpl_cmd_set_session_privilege_level_rs);
  dev->io.outofband.lan_session_state = IPMI_LAN_SESSION_STATE_SET_SESSION_PRIVILEGE_LEVEL;
  ERR_CLEANUP (!(ipmi_cmd_set_session_privilege_level (dev, obj_cmd_rs) < 0));

  dev->io.outofband.lan_session_state = IPMI_LAN_SESSION_STATE_IN_SESSION;
  rv = 0;
 cleanup:
  FIID_OBJ_DESTROY_NO_RETURN(obj_cmd_rs);
  return (rv);
}
