/* 
   ipmi-udm.c: IPMI Unified Driver Model (API interface for all
   IPMI drivers)

   Copyright (C) 2005 FreeIPMI Core Team

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  

*/

#if HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#ifdef STDC_HEADERS
#include <string.h>
#endif /* STDC_HEADERS */
#if HAVE_UNISTD_H
#include <unistd.h>
#endif /* HAVE_UNISTD_H */
#if TIME_WITH_SYS_TIME
#include <sys/time.h>
#include <time.h>
#else /* !TIME_WITH_SYS_TIME */
#if HAVE_SYS_TIME_H
#include <sys/time.h>
#else /* !HAVE_SYS_TIME_H */
#include <time.h>
#endif /* !HAVE_SYS_TIME_H */
#endif  /* !TIME_WITH_SYS_TIME */
#include <netdb.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <errno.h>

#include "freeipmi/udm/ipmi-udm.h"
#include "freeipmi/ipmi-debug.h"
#include "freeipmi/ipmi-locate.h"
#include "freeipmi/ipmi-kcs-api.h"
#include "freeipmi/ipmi-lan-interface.h"
#include "freeipmi/ipmi-openipmi-api.h"
#include "freeipmi/ipmi-ssif-api.h"
#include "freeipmi/ipmi-utils.h"
#include "freeipmi/rmcp.h"
#include "freeipmi/udm/ipmi-kcs-api-udm.h"
#include "freeipmi/udm/ipmi-lan-interface-udm.h"
#include "freeipmi/udm/ipmi-messaging-support-cmds-udm.h"
#include "freeipmi/udm/ipmi-openipmi-api-udm.h"
#include "freeipmi/udm/ipmi-ssif-api-udm.h"

#include "ipmi-udm-device.h"

#include "err-wrappers.h"
#include "fiid-wrappers.h"
#include "freeipmi-portability.h"
#include "ipmi-common.h"

#define IPMI_SESSION_TIMEOUT     60000
#define IPMI_RETRY_TIMEOUT       1000
#define IPMI_POLL_INTERVAL_USECS 60

#define GETHOSTBYNAME_AUX_BUFLEN 1024
extern int h_errno;

static ipmi_device_t
_ipmi_device_create(void)
{
  struct ipmi_device *dev;

  if (!(dev = (struct ipmi_device *)malloc(sizeof(struct ipmi_device))))
    return NULL;

  memset(dev, '\0', sizeof(struct ipmi_device));
  dev->magic = IPMI_UDM_DEVICE_MAGIC;

  return dev;
}

static void
_ipmi_device_destroy(ipmi_device_t dev)
{
  ERR_EINVAL_VOID_RETURN(dev && dev->magic == IPMI_UDM_DEVICE_MAGIC);

  free(dev);
}

static void 
_ipmi_outofband_free (ipmi_device_t dev)
{
  ERR_EINVAL_VOID_RETURN(dev && dev->magic == IPMI_UDM_DEVICE_MAGIC);

  FIID_OBJ_DESTROY_NO_RETURN (dev->io.outofband.rq.obj_rmcp_hdr);
  FIID_OBJ_DESTROY_NO_RETURN (dev->io.outofband.rs.obj_rmcp_hdr);
  FIID_OBJ_DESTROY_NO_RETURN (dev->io.outofband.rq.obj_lan_session_hdr);
  FIID_OBJ_DESTROY_NO_RETURN (dev->io.outofband.rs.obj_lan_session_hdr);
  FIID_OBJ_DESTROY_NO_RETURN (dev->io.outofband.rq.obj_lan_msg_hdr);
  FIID_OBJ_DESTROY_NO_RETURN (dev->io.outofband.rs.obj_lan_msg_hdr);
  FIID_OBJ_DESTROY_NO_RETURN (dev->io.outofband.rq.obj_lan_msg_trlr);
  FIID_OBJ_DESTROY_NO_RETURN (dev->io.outofband.rs.obj_lan_msg_trlr);
}

static void 
_ipmi_inband_free (ipmi_device_t dev)
{
  ERR_EINVAL_VOID_RETURN(dev && dev->magic == IPMI_UDM_DEVICE_MAGIC);
  
  if (dev->type == IPMI_DEVICE_KCS && dev->io.inband.kcs_ctx)
    ipmi_kcs_ctx_destroy(dev->io.inband.kcs_ctx);
  if (dev->type == IPMI_DEVICE_SSIF && dev->io.inband.ssif_ctx)
    ipmi_ssif_ctx_destroy(dev->io.inband.ssif_ctx);
  if (dev->type == IPMI_DEVICE_OPENIPMI && dev->io.inband.openipmi_ctx)
    ipmi_openipmi_ctx_destroy(dev->io.inband.openipmi_ctx);

  FIID_OBJ_DESTROY_NO_RETURN (dev->io.inband.rq.obj_hdr);
  FIID_OBJ_DESTROY_NO_RETURN (dev->io.inband.rs.obj_hdr);
}

ipmi_device_t
ipmi_open_outofband (ipmi_driver_type_t driver_type, 
		     char *hostname,
		     char *username, 
		     char *password, 
		     uint8_t authentication_type, 
		     uint8_t privilege_level,
                     unsigned int session_timeout,
                     unsigned int retry_timeout,
                     uint32_t flags)
{
  struct sockaddr_in addr;
  ipmi_device_t dev;
#ifdef HAVE_FUNC_GETHOSTBYNAME_R_6
  struct hostent hent;
  int h_errnop;
  char buf[GETHOSTBYNAME_AUX_BUFLEN];
#endif /* HAVE_FUNC_GETHOSTBYNAME_R_6 */
  struct hostent *hptr;

  ERR_EINVAL_NULL_RETURN (driver_type == IPMI_DEVICE_LAN
			  && hostname
                          && !(username && strlen (username) > IPMI_MAX_USER_NAME_LENGTH)
                          && !(password && strlen (password) > IPMI_MAX_AUTHENTICATION_CODE_LENGTH)
                          && IPMI_1_5_AUTHENTICATION_TYPE_VALID (authentication_type)
                          && IPMI_PRIVILEGE_LEVEL_VALID (privilege_level));
  
  if (!(dev = _ipmi_device_create()))
    goto cleanup;

  dev->type = driver_type;
  dev->flags = flags;

#ifdef HAVE_FUNC_GETHOSTBYNAME_R_6
  memset(&hent, '\0', sizeof(struct hostent));
  ERR_EINVAL_CLEANUP (!gethostbyname_r(hostname,
				       &hent,
				       buf,
				       GETHOSTBYNAME_AUX_BUFLEN,
				       &hptr,
				       &h_errnop));
  ERR_EINVAL_CLEANUP (hptr);
#else  /* !HAVE_FUNC_GETHOSTBYNAME_R */
#error Additional threadsafe gethostbyname support needed
#endif /* !HAVE_FUNC_GETHOSTBYNAME_R */

  dev->io.outofband.remote_host.sin_family = AF_INET;
  dev->io.outofband.remote_host.sin_port = htons(RMCP_AUX_BUS_SHUNT);
  dev->io.outofband.remote_host.sin_addr = *(struct in_addr *) hptr->h_addr;

  dev->io.outofband.session_timeout = (session_timeout ? session_timeout : IPMI_SESSION_TIMEOUT);
  dev->io.outofband.retry_timeout = (retry_timeout ? retry_timeout : IPMI_RETRY_TIMEOUT);
  memset(&dev->io.outofband.last_received, '\0', sizeof(struct timeval));
  dev->io.outofband.authentication_type = authentication_type;
  memset(dev->io.outofband.username, '\0', IPMI_MAX_USER_NAME_LENGTH);
  if (username != NULL)
    memcpy (dev->io.outofband.username, 
	    username, 
	    strlen (username));
  memset(dev->io.outofband.password, '\0', IPMI_MAX_AUTHENTICATION_CODE_LENGTH);
  if (password != NULL)
    memcpy (dev->io.outofband.password, 
	    password, 
	    strlen (password));
  dev->io.outofband.privilege_level = privilege_level;

  /* achu: per_msg_auth_disabled and lan_session_state used and
   * initialized in ipmi_lan_open_session()
   */
  
  FIID_OBJ_CREATE_CLEANUP (dev->io.outofband.rq.obj_rmcp_hdr, tmpl_rmcp_hdr);
  
  FIID_OBJ_CREATE_CLEANUP (dev->io.outofband.rs.obj_rmcp_hdr, tmpl_rmcp_hdr);
  
  FIID_OBJ_CREATE_CLEANUP (dev->io.outofband.rq.obj_lan_session_hdr, tmpl_lan_session_hdr);
  
  FIID_OBJ_CREATE_CLEANUP (dev->io.outofband.rs.obj_lan_session_hdr, tmpl_lan_session_hdr);
  
  FIID_OBJ_CREATE_CLEANUP (dev->io.outofband.rq.obj_lan_msg_hdr, tmpl_lan_msg_hdr_rq);
  
  FIID_OBJ_CREATE_CLEANUP (dev->io.outofband.rs.obj_lan_msg_hdr, tmpl_lan_msg_hdr_rs);
  
  FIID_OBJ_CREATE_CLEANUP (dev->io.outofband.rq.obj_lan_msg_trlr, tmpl_lan_msg_trlr);
  
  FIID_OBJ_CREATE_CLEANUP (dev->io.outofband.rs.obj_lan_msg_trlr, tmpl_lan_msg_trlr);
  
  /* Open client (local) UDP socket */
  /* achu: ephemeral ports are > 1023, so no way we will bind to a IPMI port */
  
  ERR_CLEANUP (!((dev->io.outofband.local_sockfd = socket (AF_INET, SOCK_DGRAM, 0)) < 0));

  memset (&addr, 0, sizeof (struct sockaddr_in));
  addr.sin_family = AF_INET;
  addr.sin_port   = htons (0);
  addr.sin_addr.s_addr = htonl (INADDR_ANY);

  ERR_CLEANUP (!(bind(dev->io.outofband.local_sockfd, 
                      (struct sockaddr *)&addr,
                      sizeof(struct sockaddr_in)) < 0));

  /* Note that ipmi_lan_open_session itself calls ipmi_lan_cmd many
     times internally, at this point everything must be set to go
     -- Anand Babu */
  ERR_CLEANUP (!(ipmi_lan_open_session (dev) < 0));

  return (dev);

 cleanup:
  if (dev->io.outofband.local_sockfd)
    close (dev->io.outofband.local_sockfd);
  if (dev)
    {
      _ipmi_outofband_free (dev);
      _ipmi_device_destroy (dev);
    }
  return (NULL);
}

ipmi_device_t
ipmi_open_inband (ipmi_driver_type_t driver_type, 
		  int disable_auto_probe, 
		  uint16_t driver_address, 
		  uint8_t reg_space,
		  char *driver_device, 
		  uint32_t flags)
{
  struct ipmi_locate_info locate_info;
  ipmi_device_t dev;
  uint32_t temp_flags = 0;

  ERR_EINVAL_NULL_RETURN (driver_type == IPMI_DEVICE_KCS
                          || driver_type == IPMI_DEVICE_SMIC
                          || driver_type == IPMI_DEVICE_BT
                          || driver_type == IPMI_DEVICE_SSIF
                          || driver_type == IPMI_DEVICE_OPENIPMI);

  if (!(dev = _ipmi_device_create()))
    goto cleanup;

  dev->io.inband.kcs_ctx = NULL;
  dev->io.inband.ssif_ctx = NULL;
  dev->io.inband.openipmi_ctx = NULL;

  switch (driver_type)
    {
    case IPMI_DEVICE_KCS:
      if (disable_auto_probe)
	{
	  memset(&locate_info, '\0', sizeof(struct ipmi_locate_info));

	  locate_info.ipmi_ver_major = 1;
	  locate_info.ipmi_ver_minor = 5;
	  locate_info.locate_driver_type = IPMI_LOCATE_DRIVER_NONE;
	  locate_info.interface_type = IPMI_INTERFACE_KCS;
	  if (driver_device)
	    {
	      strncpy(locate_info.driver_device, driver_device, IPMI_LOCATE_PATH_MAX);
	      locate_info.driver_device[IPMI_LOCATE_PATH_MAX - 1] = '\0';
	    }
	  locate_info.address_space_id = IPMI_ADDRESS_SPACE_ID_SYSTEM_IO;
	  locate_info.driver_address = driver_address;
	  locate_info.reg_space = reg_space;
	}
      else 
	{
	  ERR_CLEANUP (ipmi_locate (IPMI_INTERFACE_KCS, &locate_info) == 0);
	  if (driver_device)
	    {
	      strncpy(locate_info.driver_device, driver_device, IPMI_LOCATE_PATH_MAX);
	      locate_info.driver_device[IPMI_LOCATE_PATH_MAX - 1] = '\0';
	    }
	  if (driver_address)
	    locate_info.driver_address = driver_address;
	  if (reg_space)
	    locate_info.reg_space = reg_space;
	}
      dev->type = driver_type;
      dev->flags = flags;
      
      /* At this point we only support SYSTEM_IO, i.e. inb/outb style IO. 
	 If we cant find the bass address, we better exit. -- Anand Babu */
      ERR_ENODEV_CLEANUP (locate_info.address_space_id == IPMI_ADDRESS_SPACE_ID_SYSTEM_IO);

      ERR_CLEANUP ((dev->io.inband.kcs_ctx = ipmi_kcs_ctx_create()));
      
      ERR_CLEANUP (!(ipmi_kcs_ctx_set_driver_address(dev->io.inband.kcs_ctx, 
						     locate_info.driver_address) < 0));

      ERR_CLEANUP (!(ipmi_kcs_ctx_set_register_space(dev->io.inband.kcs_ctx, 
						     locate_info.reg_space) < 0));

      ERR_CLEANUP (!(ipmi_kcs_ctx_set_poll_interval(dev->io.inband.kcs_ctx, 
						    IPMI_POLL_INTERVAL_USECS) < 0));

      if (dev->flags & IPMI_FLAGS_NONBLOCKING)
        temp_flags |= IPMI_KCS_FLAGS_NONBLOCKING;
      
      ERR_CLEANUP (!(ipmi_kcs_ctx_set_flags(dev->io.inband.kcs_ctx, temp_flags) < 0));

      ERR_CLEANUP (!(ipmi_kcs_ctx_io_init(dev->io.inband.kcs_ctx) < 0));

      break;
    case IPMI_DEVICE_SMIC:
      errno = ENOTSUP;
      goto cleanup;
    case IPMI_DEVICE_BT:
      errno = ENOTSUP;
      goto cleanup;
    case IPMI_DEVICE_SSIF:
      if (disable_auto_probe)
	{
	  ERR_EINVAL_CLEANUP (driver_device);

	  memset(&locate_info, '\0', sizeof(struct ipmi_locate_info));

	  locate_info.ipmi_ver_major = 1;
	  locate_info.ipmi_ver_minor = 5;
	  locate_info.locate_driver_type = IPMI_LOCATE_DRIVER_NONE;
	  locate_info.interface_type = IPMI_INTERFACE_SSIF;
	  strncpy(locate_info.driver_device, driver_device, IPMI_LOCATE_PATH_MAX);
	  locate_info.driver_device[IPMI_LOCATE_PATH_MAX - 1] = '\0';
	  locate_info.address_space_id = IPMI_ADDRESS_SPACE_ID_SMBUS;
	  locate_info.driver_address = driver_address;
	  locate_info.reg_space = reg_space;
	}
      else 
	{
	  ERR_CLEANUP (ipmi_locate (IPMI_INTERFACE_SSIF, &locate_info) == 0);
	  if (driver_device)
	    {
	      strncpy(locate_info.driver_device, driver_device, IPMI_LOCATE_PATH_MAX);
	      locate_info.driver_device[IPMI_LOCATE_PATH_MAX - 1] = '\0';
	    }
	  if (driver_address)
	    locate_info.driver_address = driver_address;
	  if (reg_space)
	    locate_info.reg_space = reg_space;
	}
      dev->type = driver_type;
      dev->flags = flags;

      ERR_CLEANUP ((dev->io.inband.ssif_ctx = ipmi_ssif_ctx_create()));
      
      ERR_CLEANUP (!(ipmi_ssif_ctx_set_driver_device(dev->io.inband.ssif_ctx, 
						     locate_info.driver_device) < 0));
 
      ERR_CLEANUP (!(ipmi_ssif_ctx_set_driver_address(dev->io.inband.ssif_ctx, 
						      locate_info.driver_address) < 0));

      if (dev->flags & IPMI_FLAGS_NONBLOCKING)
        temp_flags |= IPMI_SSIF_FLAGS_NONBLOCKING;
      
      ERR_CLEANUP (!(ipmi_ssif_ctx_set_flags(dev->io.inband.ssif_ctx, temp_flags) < 0));

      ERR_CLEANUP (!(ipmi_ssif_ctx_io_init(dev->io.inband.ssif_ctx) < 0));

      break;

    case IPMI_DEVICE_OPENIPMI:
      dev->type = driver_type;
      dev->flags = flags;

      ERR_CLEANUP ((dev->io.inband.openipmi_ctx = ipmi_openipmi_ctx_create()));
      
      if (driver_device)
        ERR_CLEANUP (!(ipmi_openipmi_ctx_set_driver_device(dev->io.inband.openipmi_ctx,
							   driver_device) < 0));

      ERR_CLEANUP (!(ipmi_openipmi_ctx_io_init(dev->io.inband.openipmi_ctx) < 0));

      break;

    default:
      goto cleanup;
    }
  
  /* Prepare in-band headers */
  FIID_OBJ_CREATE_CLEANUP (dev->io.inband.rq.obj_hdr, tmpl_hdr_kcs);
  FIID_OBJ_CREATE_CLEANUP (dev->io.inband.rs.obj_hdr, tmpl_hdr_kcs);
  
  return (dev);

 cleanup:
  if (dev)
    {
      _ipmi_inband_free (dev);
      _ipmi_device_destroy (dev);
    }
  return (NULL);
}

int 
ipmi_cmd (ipmi_device_t dev, 
	  uint8_t lun, 
	  uint8_t net_fn, 
	  fiid_obj_t obj_cmd_rq, 
	  fiid_obj_t obj_cmd_rs)
{
  int8_t status = 0;

  ERR_EINVAL (dev 
              && dev->magic == IPMI_UDM_DEVICE_MAGIC
	      && (dev->type == IPMI_DEVICE_LAN
		  || dev->type == IPMI_DEVICE_KCS
		  || dev->type == IPMI_DEVICE_SSIF
                  || dev->type == IPMI_DEVICE_OPENIPMI));

  FIID_OBJ_PACKET_VALID(obj_cmd_rq);

  dev->lun = lun;
  dev->net_fn = net_fn;
  
  if (dev->flags & IPMI_FLAGS_DEBUG_DUMP)
    {
      /* lan packets are dumped in ipmi lan code */
      if (dev->type != IPMI_DEVICE_LAN)
	ipmi_obj_dump(STDERR_FILENO, obj_cmd_rq);
    }

  switch (dev->type)
    {
    case IPMI_DEVICE_LAN:
      status = ipmi_lan_cmd (dev, obj_cmd_rq, obj_cmd_rs);
      break;
    case IPMI_DEVICE_KCS:
      status = ipmi_kcs_cmd_udm (dev, obj_cmd_rq, obj_cmd_rs);
      break;
    case IPMI_DEVICE_SSIF:
      status = ipmi_ssif_cmd_udm (dev, obj_cmd_rq, obj_cmd_rs);
      break;
    case IPMI_DEVICE_OPENIPMI:
      status = ipmi_openipmi_cmd_udm (dev, obj_cmd_rq, obj_cmd_rs);
      break;
    default:
      ERR_EXIT(0);
    }
  
  if (dev->flags & IPMI_FLAGS_DEBUG_DUMP)
    {
      /* lan packets are dumped in ipmi lan code */
      if (dev->type != IPMI_DEVICE_LAN)
	ipmi_obj_dump(STDERR_FILENO, obj_cmd_rs);
    }

  return (status);
}

int 
ipmi_cmd_raw (ipmi_device_t dev, 
              uint8_t lun,
              uint8_t net_fn,
	      uint8_t *in, 
	      size_t in_len, 
	      uint8_t *out, 
	      size_t out_len)
{
  int8_t status = 0;

  ERR_EINVAL(dev
             && dev->magic == IPMI_UDM_DEVICE_MAGIC
	     && in_len > 0
	     && (dev->type == IPMI_DEVICE_LAN
		 || dev->type == IPMI_DEVICE_KCS
		 || dev->type == IPMI_DEVICE_SSIF
                 || dev->type == IPMI_DEVICE_OPENIPMI));
  
  dev->lun = lun;
  dev->net_fn = net_fn;

  switch (dev->type)
    {
    case IPMI_DEVICE_LAN:
      status = ipmi_lan_cmd_raw (dev, in, in_len, out, out_len);
      break;
    case IPMI_DEVICE_KCS:
      status = ipmi_kcs_cmd_raw_udm (dev, in, in_len, out, out_len);
      break;
    case IPMI_DEVICE_SSIF:
      status = ipmi_ssif_cmd_raw_udm (dev, in, in_len, out, out_len);
      break;
    case IPMI_DEVICE_OPENIPMI:
      status = ipmi_openipmi_cmd_raw_udm (dev, in, in_len, out, out_len);
      break;
    default:
      ERR_EXIT(0);
    }
  return (status);
}

static void
_ipmi_outofband_close (ipmi_device_t dev)
{
  fiid_obj_t obj_cmd_rs = NULL;
  int retval;

  ERR_EINVAL_VOID_RETURN(dev && dev->magic == IPMI_UDM_DEVICE_MAGIC);
 
  FIID_OBJ_CREATE_CLEANUP (obj_cmd_rs, tmpl_cmd_close_session_rs);

  dev->io.outofband.lan_session_state = IPMI_LAN_SESSION_STATE_CLOSE_SESSION;
  ERR_CLEANUP (!((retval = ipmi_lan_close_session (dev, obj_cmd_rs)) < 0));

  if (dev->io.outofband.local_sockfd)
    close (dev->io.outofband.local_sockfd);
  
  _ipmi_outofband_free (dev);

cleanup:
  FIID_OBJ_DESTROY_NO_RETURN(obj_cmd_rs);
}

static void
_ipmi_inband_close (ipmi_device_t dev)
{
  ERR_EINVAL_VOID_RETURN(dev 
                         && dev->magic == IPMI_UDM_DEVICE_MAGIC
                         && (dev->type == IPMI_DEVICE_KCS
                             || dev->type == IPMI_DEVICE_SMIC
                             || dev->type == IPMI_DEVICE_BT
                             || dev->type == IPMI_DEVICE_SSIF
                             || dev->type == IPMI_DEVICE_OPENIPMI));
                         
  _ipmi_inband_free (dev);
}

void
ipmi_close_device (ipmi_device_t dev)
{
  ERR_EINVAL_VOID_RETURN (dev
                          && dev->magic == IPMI_UDM_DEVICE_MAGIC
                          && (dev->type == IPMI_DEVICE_LAN
                              || dev->type == IPMI_DEVICE_KCS
                              || dev->type == IPMI_DEVICE_SMIC
                              || dev->type == IPMI_DEVICE_BT
                              || dev->type == IPMI_DEVICE_SSIF
                              || dev->type == IPMI_DEVICE_OPENIPMI));
  
  switch (dev->type)
    {
    case IPMI_DEVICE_LAN:
      _ipmi_outofband_close (dev);
      break;
    case IPMI_DEVICE_KCS:
    case IPMI_DEVICE_SMIC:
    case IPMI_DEVICE_BT:
    case IPMI_DEVICE_SSIF:
    case IPMI_DEVICE_OPENIPMI:
      _ipmi_inband_close (dev);
      break;
    default:
      break;
    }

  _ipmi_device_destroy(dev);
}
